"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __param = (this && this.__param) || function (paramIndex, decorator) {
    return function (target, key) { decorator(target, key, paramIndex); }
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const decorators_1 = require("../decorators");
const Server_1 = require("../Server");
const Token_1 = require("../shared/Token");
const realms_1 = require("../realms");
const errors = require("../errors");
const path = require("path");
const moment = require("moment");
const URI = require("urijs");
const _ = require("lodash");
const events_1 = require("events");
let PermissionService = PermissionService_1 = class PermissionService extends events_1.EventEmitter {
    get adminRealm() {
        return this._syncedAdminRealm.realm;
    }
    get starPermissionsRealm() {
        return this._syncedStarPermissionsRealm.realm;
    }
    serverStarted(server) {
        this.server = server;
        this.dataPath = path.resolve(server.config.dataPath, "permissions");
        this.privateKey = server.privateKey;
        this.publicKey = server.publicKey;
        this.logger = server.logger;
        this.adminTokenUser = Realm.Sync.User.adminUser(Token_1.generateAdminToken(this.privateKey), `http://${server.address}`);
        this._syncedAdminRealm = server.getSyncedRealm(realms_1.AdminRealm);
        this._syncedStarPermissionsRealm = server.getSyncedRealm(realms_1.StarPermissionRealm);
        const syncBaseUrl = `realm://${server.address}`;
        const authUrl = `http://${server.address}`;
        this.handleAdminChangeCallback = this.handleAdminRealmPermissionToReflection.bind(this);
        Realm.Sync.addListener(syncBaseUrl, Realm.Sync.User.adminUser(Token_1.generateAdminToken(this.privateKey), authUrl), PermissionService_1.ADMIN_REALM_REGEX, 'change', this.handleAdminChangeCallback);
        this.handleManagementChangeCallback = this.handleManagementRealmChange.bind(this);
        Realm.Sync.addListener(syncBaseUrl, Realm.Sync.User.adminUser(Token_1.generateAdminToken(this.privateKey), authUrl), PermissionService_1.MANAGEMENT_REALM_REGEX, 'change', this.handleManagementChangeCallback);
    }
    stop() {
        Realm.Sync.removeListener(PermissionService_1.MANAGEMENT_REALM_REGEX, 'change', this.handleManagementChangeCallback);
        Realm.Sync.removeListener(PermissionService_1.ADMIN_REALM_REGEX, 'change', this.handleAdminRealmPermissionToReflection);
    }
    onDemand(accessToken) {
        if (!accessToken) {
            throw new errors.JSONError({
                title: "X-Realm-Access-Token header is not valid",
                status: 400,
            });
        }
        const token = Token_1.Token.parse(accessToken, this.publicKey);
        const userId = token.identity;
        const permissionsInAdminRealm = this.adminRealm.objects('Permission')
            .filtered('user.userId = $0 || realmFile.owner.userId = $0', userId);
        const userSpecificRealm = this.getUserSpecificPermissionRealm(userId);
        let didUpdateAnyPermission = false;
        userSpecificRealm.beginTransaction();
        for (let permissionFromAdminRealm of permissionsInAdminRealm.snapshot()) {
            didUpdateAnyPermission = this.updateOrCreateUserPermissionWithAdminPermission(userSpecificRealm, permissionFromAdminRealm);
        }
        if (didUpdateAnyPermission) {
            userSpecificRealm.commitTransaction();
        }
        else {
            userSpecificRealm.cancelTransaction();
        }
        return {
            userId: userId,
            affectedPermissionsCount: permissionsInAdminRealm.length
        };
    }
    updateOrCreateUserPermissionWithAdminPermission(userSpecificRealm, adminPermission) {
        const userSpecificPermission = userSpecificRealm.objects("Permission")
            .filtered('userId = $0 && path = $1', adminPermission.user.userId, adminPermission.realmFile.path)[0];
        if (!userSpecificPermission) {
            userSpecificRealm.create(realms_1.UserSpecificPermission.schema.name, {
                path: adminPermission.realmFile.path,
                userId: adminPermission.user.userId,
                mayRead: adminPermission.mayRead,
                mayWrite: adminPermission.mayWrite,
                mayManage: adminPermission.mayManage,
                updatedAt: new Date()
            });
            return true;
        }
        let updatePermission = (userSpecificPermission.mayRead !== adminPermission.mayRead) ||
            (userSpecificPermission.mayWrite !== adminPermission.mayWrite) ||
            (userSpecificPermission.mayManage !== adminPermission.mayManage);
        if (updatePermission) {
            userSpecificPermission.mayRead = adminPermission.mayRead;
            userSpecificPermission.mayWrite = adminPermission.mayWrite;
            userSpecificPermission.mayManage = adminPermission.mayManage;
            userSpecificPermission.updatedAt = new Date();
            return true;
        }
        return false;
    }
    getUserByUserId(userId) {
        return this.adminRealm.objectForPrimaryKey('User', userId);
    }
    getAllUsers() {
        return this.adminRealm.objects('User');
    }
    getUserIdsWithMetadata(key, value) {
        let metadata = this.adminRealm.objects('UserMetadataRow');
        if (key) {
            metadata = metadata.filtered("key = $0", key);
        }
        if (value) {
            metadata = metadata.filtered("value = $0", value);
        }
        return Array.from(new Set(metadata.map((meta) => meta.user[0].userId)));
    }
    getUsersByUserIds(userIds) {
        const query = userIds.map((id, index) => {
            return `userId = $${index}`;
        }).join(" || ");
        return this.adminRealm.objects("User").filtered(query, ...userIds);
    }
    getManagementRealmForUserId(userId) {
        const realmDefinition = realms_1.CreateUserManagementRealmDefinition(userId);
        const syncedRealm = this.server.getSyncedRealm(realmDefinition);
        return syncedRealm.realm;
    }
    applyPermissionChangeInAdminRealm(realmFilePath, userId, mayRead, mayWrite, mayManage) {
        const realmFile = this.adminRealm.objectForPrimaryKey("RealmFile", realmFilePath);
        if (!realmFile) {
            throw new errors.realm.InvalidParameters('realmFilePath does not exist');
        }
        let permission;
        let user;
        this.adminRealm.write(() => {
            if (userId) {
                user = this.getUserByUserId(userId);
                permission = realmFile.permissions.filtered('user.userId = $0', userId)[0];
            }
            else {
                permission = realmFile.permissions.filtered('user == null')[0];
            }
            if (!permission) {
                permission = this.adminRealm.create('Permission', {
                    user: user,
                    realmFile: realmFile,
                    updatedAt: new Date(),
                    mayRead: mayRead || false,
                    mayWrite: mayWrite || false,
                    mayManage: mayManage || false
                });
            }
            else {
                permission.mayRead = mayRead || permission.mayRead;
                permission.mayWrite = mayWrite || permission.mayWrite;
                permission.mayManage = mayManage || permission.mayManage;
                permission.updatedAt = new Date();
            }
        });
    }
    getUserSpecificPermissionRealm(userId) {
        const realmDefinition = realms_1.CreateUserPermissionRealmDefinition(userId);
        const syncedUserRealm = this.server.getSyncedRealm(realmDefinition);
        return syncedUserRealm.realm;
    }
    getUserSpecificPermissionAndAssociatedRealm(realmPath, userId) {
        const realmDefinition = realms_1.CreateUserPermissionRealmDefinition(userId);
        const syncedUserRealm = this.server.getSyncedRealm(realmDefinition);
        const userSpecificPermission = syncedUserRealm.realm.objects("Permission").filtered('path = $0', realmPath)[0];
        return {
            userSpecificRealm: syncedUserRealm.realm,
            userSpecificPermission: userSpecificPermission
        };
    }
    handleAdminRealmPermissionToReflection(changeEvent) {
        return __awaiter(this, void 0, void 0, function* () {
            const permissionChanges = changeEvent.changes['Permission'];
            if (permissionChanges) {
                const collection = changeEvent.realm.objects('Permission');
                this.emit('handlePermissionReflection', changeEvent);
                const insertedPermissions = permissionChanges.insertions.map(i => collection[i]);
                const modifiedPermissions = permissionChanges.modifications.map(i => collection[i]);
                const insertedAndChangedPermission = insertedPermissions.concat(modifiedPermissions);
                for (let permission of insertedAndChangedPermission) {
                    if (!permission.user) {
                        continue;
                    }
                    const { userSpecificRealm, userSpecificPermission } = this.getUserSpecificPermissionAndAssociatedRealm(permission.realmFile.path, permission.user.userId);
                    userSpecificRealm.write(() => {
                        if (!userSpecificPermission) {
                            userSpecificRealm.create('Permission', {
                                mayRead: permission.mayRead,
                                mayWrite: permission.mayWrite,
                                mayManage: permission.mayManage,
                                path: permission.realmFile.path,
                                userId: permission.user.userId,
                                updatedAt: new Date()
                            });
                        }
                        else {
                            userSpecificPermission.mayRead = permission.mayRead;
                            userSpecificPermission.mayWrite = permission.mayWrite;
                            userSpecificPermission.mayManage = permission.mayManage;
                            userSpecificPermission.path = permission.realmFile.path;
                            userSpecificPermission.updatedAt = new Date();
                            userSpecificPermission.userId = permission.user.userId;
                        }
                    });
                }
                if (permissionChanges.deletions.length > 0) {
                    const oldCollection = changeEvent.oldRealm.objects('Permission');
                    const deletedPermissions = permissionChanges.deletions.map(i => oldCollection[i]);
                    for (let permission of deletedPermissions) {
                        const { userSpecificRealm, userSpecificPermission } = this.getUserSpecificPermissionAndAssociatedRealm(permission.realmFile.path, permission.user.userId);
                        if (!userSpecificPermission) {
                        }
                        else {
                            userSpecificRealm.write(() => {
                                userSpecificRealm.delete(userSpecificPermission);
                            });
                        }
                    }
                }
            }
        });
    }
    handleManagementRealmChange(changeEvent) {
        return __awaiter(this, void 0, void 0, function* () {
            const matches = changeEvent.path.match(PermissionService_1.MANAGEMENT_REALM_REGEX);
            if (!matches) {
                this.logger.error(`Received changeEvent for non-management realm: ${changeEvent.path}`);
                return;
            }
            const userId = matches[1];
            const user = this.getUserByUserId(userId);
            const realmPath = changeEvent.path;
            const realm = changeEvent.realm;
            this.logger.debug(`A management realm with path ${changeEvent.path} has changes with userId: ${userId}`);
            if (!user) {
                this.logger.error(`A change notification with management realm of path ${changeEvent.path}
            was notified with a userId of "${userId}" but this userId was not found in the system. This realm will now be closed.
            Please ensure that ${userId} exists within the __admin Realm. This permission change attempt will be ignored.`);
                realm.close();
                return;
            }
            if (realm.schema.findIndex(objSchema => { return objSchema.name === 'PermissionOffer'; }) !== -1) {
                let unprocessedPermissionOfferObjects = realm.objects("PermissionOffer").filtered('statusCode == null');
                for (let p of unprocessedPermissionOfferObjects.snapshot()) {
                    this.handlePermissionOfferObject(p, userId, realm);
                }
            }
            if (realm.schema.findIndex(objSchema => { return objSchema.name === 'PermissionOfferResponse'; }) !== -1) {
                let unprocessedPermissionOfferResponseObjects = realm.objects("PermissionOfferResponse").filtered('statusCode == null');
                for (let p of unprocessedPermissionOfferResponseObjects.snapshot()) {
                    yield this.handlePermissionOfferResponseObject(p, userId, realm);
                }
            }
            if (realm.schema.findIndex(objSchema => { return objSchema.name === 'PermissionChange'; }) !== -1) {
                let unprocessedPermissionChangeObjects = realm.objects("PermissionChange").filtered('statusCode == null');
                for (let p of unprocessedPermissionChangeObjects.snapshot()) {
                    this.handlePermissionChangeObject(p, userId, realm);
                }
            }
        });
    }
    handlePermissionOfferObject(permissionOffer, userId, realm) {
        try {
            const realmPath = new URI(permissionOffer.realmUrl).path().replace('~', userId);
            this.validateUserIdCanChangePermissionsForPath(userId, realmPath);
            if (permissionOffer.expiresAt && moment(permissionOffer.expiresAt).isBefore(moment.now())) {
                throw new errors.realm.ExpiredPermissionOffer({
                    title: `The permission offer has expired on ${permissionOffer.expiresAt}. The current time is ${new Date()}.`
                });
            }
            realm.write(() => {
                permissionOffer.realmUrl = realmPath;
                permissionOffer.token = `${userId}:${permissionOffer.id}`;
                permissionOffer.statusCode = 0;
            });
        }
        catch (err) {
            const problem = err;
            this.logger.error(problem.toString());
            realm.write(() => {
                permissionOffer.statusMessage = problem.title;
                permissionOffer.statusCode = problem.code;
            });
        }
    }
    handlePermissionOfferResponseObject(permissionOfferResponse, userId, realm) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const tokenComponents = permissionOfferResponse.token.split(':');
                if (tokenComponents.length < 2 || tokenComponents[0].length === 0) {
                    throw new errors.realm.InvalidParameters('token');
                }
                const offeringUserId = tokenComponents[0];
                let offeringRealm = this.getManagementRealmForUserId(offeringUserId);
                while (!offeringRealm.syncSession) {
                    this.logger.debug('FIXME: offering realm is not in sync session yet, waiting');
                    yield new Promise((resolve, reject) => {
                        setTimeout(resolve, 1000);
                    });
                    offeringRealm = this.getManagementRealmForUserId(offeringUserId);
                }
                yield new Promise((resolve, reject) => {
                    const session = offeringRealm.syncSession;
                    session.addProgressNotification('download', 'forCurrentlyOutstandingWork', function listener(a, b) {
                        if (a === b) {
                            session.removeProgressNotification(listener);
                            resolve();
                        }
                    });
                });
                const permissionOffers = offeringRealm.objects("PermissionOffer").filtered('token = $0', permissionOfferResponse.token);
                if (permissionOffers.length === 0) {
                    throw new errors.realm.InvalidParameters('token');
                }
                if (permissionOffers.length > 1) {
                    throw new errors.realm.AmbiguousPermissionOfferToken({
                        detail: `Unexpected count of permission offers (=${permissionOffers.length})`
                            + ` matching the token '${permissionOfferResponse.token}' of the permission request!`,
                    });
                }
                const permissionOffer = permissionOffers[0];
                if (permissionOffer.expiresAt && moment(permissionOffer.expiresAt).isBefore(moment.now())) {
                    throw new errors.realm.ExpiredPermissionOffer({
                        title: `The permission offer has expired on ${permissionOffer.expiresAt}. The current time is ${new Date()}.`
                    });
                }
                const realmPath = new URI(permissionOffer.realmUrl).path();
                this.applyPermissionChangeInAdminRealm(realmPath, userId, permissionOffer.mayRead, permissionOffer.mayWrite, permissionOffer.mayManage);
                realm.write(() => {
                    permissionOfferResponse.realmUrl = realmPath;
                    permissionOfferResponse.statusCode = 0;
                });
            }
            catch (err) {
                const problem = err;
                this.logger.error(problem.toString());
                realm.write(() => {
                    permissionOfferResponse.statusMessage = problem.title;
                    permissionOfferResponse.statusCode = problem.code;
                });
            }
        });
    }
    handlePermissionChangeObject(permissionChange, userId, realm) {
        try {
            const realmPath = new URI(permissionChange.realmUrl).path().replace('~', userId);
            if (permissionChange.userId === '*' && permissionChange.realmUrl === '*') {
                this.handleStarRealmUrlForAllUsers(permissionChange, userId);
            }
            else if (permissionChange.realmUrl === '*') {
                this.handleStarRealmUrlForUser(permissionChange, userId);
            }
            else if (permissionChange.userId === '*') {
                this.validateUserIdCanChangePermissionsForPath(userId, realmPath);
                this.handleStarUserIdUrl(permissionChange, realmPath);
            }
            else {
                this.validateUserIdCanChangePermissionsForPath(userId, realmPath);
                let realmFiles;
                let users;
                if (permissionChange.metadataKey) {
                    const userIds = this.getUserIdsWithMetadata(permissionChange.metadataKey, permissionChange.metadataValue);
                    this.logger.debug(`Apply permission change by metadata query 
                        (${permissionChange.metadataKey}=${permissionChange.metadataValue}) 
                        matching the userIds ${userIds}`);
                    users = this.getUsersByUserIds(userIds);
                }
                else {
                    this.logger.debug(`Apply permission change for user ${permissionChange.userId}`);
                    users = this.getUsersByUserIds([permissionChange.userId]);
                }
                if (realmPath) {
                    realmFiles = this.adminRealm.objects(realms_1.RealmFile.schema.name)
                        .filtered('path = $0', realmPath);
                }
                if (!realmFiles) {
                    realmFiles = this.adminRealm
                        .objects("RealmFile")
                        .filtered('owner.userId = $0', userId);
                }
                this.setPermissionChangesInAdminRealm(users, realmFiles, permissionChange.mayWrite, permissionChange.mayRead, permissionChange.mayManage);
            }
            realm.write(() => {
                permissionChange.statusCode = 0;
                permissionChange.statusMessage = `Successfully applied PermissionChange for ${realmPath}`;
            });
        }
        catch (err) {
            const problem = err;
            this.logger.error(problem.toString());
            realm.write(() => {
                permissionChange.statusMessage = problem.title;
                permissionChange.statusCode = problem.code;
            });
        }
    }
    validateUserIdCanChangePermissionsForPath(userId, path) {
        if (path === '') {
            throw new errors.realm.InvalidParameters({
                name: 'realmUrl',
                reason: 'The realmUrl cannot be empty and must encode a path'
            });
        }
        if (_.includes(path, '__management')) {
            throw new errors.realm.InvalidParameters({
                name: 'realmUrl',
                reason: 'The realmUrl cannot be of a management realm'
            });
        }
        if (_.includes(path, '__permission')) {
            throw new errors.realm.InvalidParameters({
                name: 'realmUrl',
                reason: 'The realmUrl cannot be of a permission realm'
            });
        }
        const user = this.getUserByUserId(userId);
        const realmFile = this.adminRealm.objectForPrimaryKey(realms_1.RealmFile.schema.name, path);
        if (!realmFile) {
            throw new errors.realm.InvalidParameters({
                name: 'realmUrl',
                reason: 'The realmUrl does not exist'
            });
        }
        const permissions = realmFile.permissions.filtered('(user.userId = $0 OR user == null) AND mayManage = true', userId);
        if (!user.isAdmin && permissions.length === 0) {
            throw new errors.realm.AccessDenied();
        }
    }
    handleStarUserIdUrl(permissionChange, path) {
        this.starPermissionsRealm.write(() => {
            this.starPermissionsRealm.create(realms_1.StarSpecificPermission.schema.name, {
                path: path,
                mayRead: permissionChange.mayRead,
                mayWrite: permissionChange.mayWrite,
                mayManage: permissionChange.mayManage,
                updatedAt: new Date()
            }, true);
        });
        this.applyPermissionChangeInAdminRealm(path, null, permissionChange.mayRead, permissionChange.mayWrite, permissionChange.mayManage);
    }
    handleStarRealmUrlForUser(permissionChange, ownerUserId) {
        const ownedRealmFiles = this.adminRealm.objects(realms_1.RealmFile.schema.name)
            .filtered('owner.userId = $0', ownerUserId);
        const users = this.getUsersByUserIds([permissionChange.userId]);
        this.setPermissionChangesInAdminRealm(users, ownedRealmFiles, permissionChange.mayWrite, permissionChange.mayRead, permissionChange.mayManage);
    }
    handleStarRealmUrlForAllUsers(permissionChange, ownerUserId) {
        const ownedRealmFiles = this.adminRealm.objects(realms_1.RealmFile.schema.name)
            .filtered('owner.userId = $0', ownerUserId);
        this.starPermissionsRealm.write(() => {
            for (let realmFile of ownedRealmFiles) {
                this.starPermissionsRealm.create(realms_1.StarSpecificPermission.schema.name, {
                    path: realmFile.path,
                    mayWrite: permissionChange.mayWrite,
                    mayManage: permissionChange.mayManage,
                    mayRead: permissionChange.mayRead,
                    updatedAt: new Date()
                }, true);
                this.applyPermissionChangeInAdminRealm(realmFile.path, null, permissionChange.mayRead, permissionChange.mayWrite, permissionChange.mayManage);
            }
        });
    }
    setPermissionChangesInAdminRealm(users, realmFiles, mayWrite, mayRead, mayManage, permissionChange) {
        for (let file of realmFiles) {
            for (let user of users) {
                this.applyPermissionChangeInAdminRealm(file.path, user.userId, mayRead, mayWrite, mayManage);
            }
        }
        if (permissionChange) {
            permissionChange.statusMessage = '';
            permissionChange.statusCode = 0;
        }
    }
};
PermissionService.MANAGEMENT_REALM_REGEX = '^/([^/]+)/__management$';
PermissionService.ADMIN_REALM_REGEX = `/__admin`;
__decorate([
    decorators_1.ServerStarted(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Server_1.Server]),
    __metadata("design:returntype", void 0)
], PermissionService.prototype, "serverStarted", null);
__decorate([
    decorators_1.Stop(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", void 0)
], PermissionService.prototype, "stop", null);
__decorate([
    decorators_1.Get('/ondemand'),
    __param(0, decorators_1.Headers('X-Realm-Access-Token')),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", void 0)
], PermissionService.prototype, "onDemand", null);
PermissionService = PermissionService_1 = __decorate([
    decorators_1.BaseRoute('/permissions'),
    decorators_1.ServiceName('permissions')
], PermissionService);
exports.PermissionService = PermissionService;
var PermissionService_1;
//# sourceMappingURL=PermissionsService.js.map